#include <config.h>
#include <stdlib.h>
#include "tools.h"
#include "io.h"
#include "data.h"
#include "parse.h"
#include "event.h"
#include "loadsave.h"
#include "command.h"


/*
 * maximum number of recognized words
 */
#define MAX_WORDS 16


/*
 * maximum number of modes than can be set at once
 */
#define MODES_AT_ONCE 3


/*
 * spaces before a log entry that belongs to a command
 */
#define CMD_INDENT "           "


/*
 * splits a string into n words logging any problems
 */
int
params_split (char **str, char *words[], int n)
{
  for (; n>0; n--, words++)
    {
      if (!parse_word (str, words))
	{
	  print2ilog (L_ERROR, CMD_INDENT "## not enough parameters");
	  return 0;
	}
    }
  return 1;
}


/*
 * checks if there are still words in str
 */
void
params_check (char *str)
{
  char *s = NULL;

  if (parse_word (&str, &s))
    {
      print2ilog (L_ERROR, CMD_INDENT "## too many parameters");
    }
}


/*
 * executes a command
 */
void
command (char *cmd)
{
  char *params[MAX_WORDS];
  char *word = NULL;
  char *param = NULL;
  char *item1 = NULL;
  char *item2 = NULL;
  char *item3 = NULL;
  char *p1, *p2, *p3;
  int i;

  for (i=0; i<MAX_WORDS; i++) params[i] = NULL;

  print2ilog (L_DETAILED, CMD_INDENT "-> %s", cmd);
  parse_word (&cmd, &word);

  /* clear <name> */
  if (strcmp (word, "clear") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  sl_clear (var_get (params[0]));
	}
      params_check (cmd);
    }

  /* add <name> <text> */
  else if (strcmp (word, "add") == 0)
    {
      struct sl *var;

      if (params_split (&cmd, params, 1))
	{
	  var = var_get (params[0]);
	  while (parse_list (&cmd, &item1))
	    {
	      sl_add_u (var, item1);
	    }
	}
    }

  /* remove <name> <text> */
  else if (strcmp (word, "remove") == 0)
    {
      struct sl *var;

      if (params_split (&cmd, params, 1))
	{
	  var = var_get (params[0]);
	  while (parse_list (&cmd, &item1))
	    {
	      sl_remove (var, item1);
	    }
	}
    }

  /* set <name> <text> */
  else if (strcmp (word, "set") == 0)
    {
      struct sl *var;

      if (params_split (&cmd, params, 1))
	{
	  var = var_get (params[0]);
	  sl_clear (var);
	  while (parse_list (&cmd, &item1))
	    {
	      sl_add_u (var, item1);
	    }
	}
    }

  /* log <file> <text> */
  else if (strcmp (word, "log") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2log (params[0], "%s", cmd);
	}
    }

  /* call <text> */
  else if (strcmp (word, "call") == 0)
    {
      char *ev = NULL;
      ev = str_cat (ev, "1001");
      ev = str_cat (ev, cmd);
      event (ev);
    }

  /* kill */
  else if (strcmp (word, "kill") == 0)
    {
      conn_close ();
      params_check (cmd);
    }

  /* update [<main file>] */
  else if (strcmp (word, "update") == 0)
    {
      if (*cmd != 000) set_main_file (cmd);
      save (sl_get (var_get ("ini"), 0));
      need_reload = 1;
    }

  /* irc <text> */
  else if (strcmp (word, "irc") == 0)
    {
      print2sock ("%s", cmd);
    }

  /* nick <nick> */
  else if (strcmp (word, "nick") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    {
	      print2sock ("NICK %s", item1);
	    }
	}
      params_check (cmd);
    }

  /* quit [<text>] */
  else if (strcmp (word, "quit") == 0)
    {
      print2sock ("QUIT :%s", cmd);
    }

  /* join <channel> */
  else if (strcmp (word, "join") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("JOIN %s", params[0]);
	}
      params_check (cmd);
    }

  /* part <channel> [<text>] */
  else if (strcmp (word, "part") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("PART %s :%s", params[0], cmd);
	}
    }

  /* kick <channel> <to nick> [<text>] */
  else if (strcmp (word, "kick") == 0)
    {
      if (params_split (&cmd, params, 2))
	{
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    for (p2 = params[1]; parse_list (&p2, &item2);)
	      {
		print2sock ("KICK %s %s :%s", item1, item2, cmd);
	      }
	}
    }

  /* msg <channel/nick> [<text>] */
  else if (strcmp (word, "msg") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("PRIVMSG %s :%s", params[0], cmd);
	}
    }

  /* cmsg <channel/nick> [<text>] */
  else if (strcmp (word, "cmsg") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("PRIVMSG %s :\001%s\001", params[0], cmd);
	}
    }

  /* notice <channel/nick> [<text>] */
  else if (strcmp (word, "notice") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("NOTICE %s :%s", params[0], cmd);
	}
    }

  /* notice <channel/nick> [<text>] */
  else if (strcmp (word, "cnotice") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  print2sock ("NOTICE %s :\001%s\001", params[0], cmd);
	}
    }

  /* topic <channel> [<text>] */
  else if (strcmp (word, "topic") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    {
	      print2sock ("TOPIC %s :%s", item1, cmd);
	    }
	}
    }

  /* invite <channel> <to nick> */
  else if (strcmp (word, "invite") == 0)
    {
      if (params_split (&cmd, params, 2))
	{
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    for (p2 = params[1]; parse_list (&p2, &item2);)
	      {
		print2sock ("INVITE %s %s", item2, item1);
	      }
	}
      params_check (cmd);
    }

  /* who <channel/mask> [<o>] */
  else if (strcmp (word, "who") == 0)
    {
      if (params_split (&cmd, params, 1))
	{
	  char *o = parse_word (&cmd, &item1) ? " o" : "";
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    {
	      print2sock ("WHO %s%s", item1, o);
	    }
	}
      params_check (cmd);
    }

  /* mode <channel/nick> <modes> [<params>] */
  else if (strcmp (word, "mode") == 0)
    {
      if (params_split (&cmd, params, 2))
	{
	  for (p1 = params[0]; parse_list (&p1, &item1);)
	    {
	      print2sock ("MODE %s %s%s%s", item1, params[1],
			  *cmd == 000 ? "" : " ", cmd);
	    }
	}
    }

  /* <+-><mode> <channel> [<params>] */
  else if (match ("+?", word) || match ("-?", word))
    {
      enum {T_0, T_1, T_LIST, T_MASK} type = T_0;
      switch (word[1])
	{
	case 'i':
	case 'm':
	case 'n':
	case 'p':
	case 's':
	case 't':
	  type = T_0;
	  break;

	case 'k':
	  type = T_1;
	  break;

	case 'o':
	case 'v':
	  type = T_LIST;
	  break;

	case 'b':
	case 'e':
	case 'I':
	  type = T_MASK;
	  break;

	case 'l':
	  type = word[0] == '+' ? T_1 : T_0;
	  break;
	}

      switch (type)
	{
	case T_0:
	  if (params_split (&cmd, params, 1))
	    {
	      for (p1 = params[0]; parse_list (&p1, &item1);)
		{
		  print2sock ("MODE %s %s", item1, word);
		}
	    }
	  break;

	case T_1:
	  if (params_split (&cmd, params, 2))
	    {
	      for (p1 = params[0]; parse_list (&p1, &item1);)
		for (p2 = params[1]; parse_list (&p2, &item2);)
		  {
		    print2sock ("MODE %s %s %s", item1, word, item2);
		  }
	    }
	  break;

	case T_LIST:
	  if (params_split (&cmd, params, 2))
	    {
	      for (p1 = params[0]; parse_list (&p1, &item1);)
		{
		  int n = 0;
		  char mode[MODES_AT_ONCE+1];
		  memset (mode, word[1], MODES_AT_ONCE);
		  mode[MODES_AT_ONCE] = 000;
		  param = str_cpy (param, "");

		  for (p2 = params[1]; parse_list (&p2, &item2);)
		    {
		      str_catchar (param, ' ');
		      str_cat (param, item2);
		      n++;
		      if (n == MODES_AT_ONCE)
			{
			  print2sock ("MODE %s %c%s%s",
				      item1, word[0], mode, param);
			  n = 0;
			  param = str_cpy (param, "");
			}
		    }

		  if (n > 0)
		    {
		      mode[n] = 000;
		      print2sock ("MODE %s %c%s%s",
				  item1, word[0], mode, param);
		    }
		}
	    }
	  break;

	case T_MASK:
	  if (params_split (&cmd, params, 3))
	    {
	      for (p1 = params[0]; parse_list (&p1, &item1);)
		{
		  int n = 0;
		  char mode[MODES_AT_ONCE+1];
		  memset (mode, word[1], MODES_AT_ONCE);
		  mode[MODES_AT_ONCE] = 000;
		  param = str_cpy (param, "");

		  for (p2 = params[1]; parse_list (&p2, &item2);)
		    for (p3 = params[2]; parse_list (&p3, &item3);)
		      {
			str_catchar (param, ' ');
			str_cat (param, item2);
			str_catchar (param, '!');
			str_cat (param, item3);
			n++;
			if (n == MODES_AT_ONCE)
			  {
			    print2sock ("MODE %s %c%s%s",
					item1, word[0], mode, param);
			    n = 0;
			    param = str_cpy (param, "");
			  }
		      }

		  if (n > 0)
		    {
		      mode[n] = 000;
		      print2sock ("MODE %s %c%s%s",
				  item1, word[0], mode, param);
		    }
		}
	    }
	  break;
	}
      params_check (cmd);
    }

  else
    {
      print2ilog (L_ERROR, CMD_INDENT "## unknown command");
    }

  free_save (word);
  free_save (param);
  free_save (item1);
  free_save (item2);
  for (i=0; i<MAX_WORDS; i++) free_save (params[i]);
}


/*
 * executes an assignment command
 */
void
assign (char *cmd)
{
  char *params[MAX_WORDS];
  char *s = NULL;
  int i;

  for (i=0; i<MAX_WORDS; i++) params[i] = NULL;

  print2ilog (L_DETAILED, CMD_INDENT "-> $%s", cmd);

  if (params_split (&cmd, params, 2))
    {
      struct sl *var = var_get (params[0]);

      /* clear */
      if (strcmp (params[1], "!") == 0)
	{
	  sl_clear (var);
	  params_check (cmd);
	}

      /* add */
      else if (strcmp (params[1], "+=") == 0)
	{
	  while (parse_list (&cmd, &s))
	    sl_add_u (var, s);
	}

      /* remove */
      else if (strcmp (params[1], "-=") == 0)
	{
	  while (parse_list (&cmd, &s))
	    sl_remove (var, s);
	}

      /* remove mask */
      else if (strcmp (params[1], "-~=") == 0)
	{
	  while (parse_list (&cmd, &s))
	    sl_remove_m (var, s);
	}

      /* set */
      else if (strcmp (params[1], "=") == 0)
	{
	  sl_clear (var);
	  while (parse_list (&cmd, &s))
	    sl_add_u (var, s);
	}

      else
	{
	  print2ilog (L_ERROR, CMD_INDENT "## unknown assignment");
	}
    }

  free_save (s);
  for (i=0; i<MAX_WORDS; i++) free_save (params[i]);
}
