#include <config.h>
#include <stdio.h>
#include <time.h>
#include "tools.h"
#include "io.h"
#include "data.h"
#include "parse.h"
#include "loadsave.h"
#include "command.h"
#include "event.h"


/*
 * says whether a reload is needed
 */
int need_reload = 0;


/*
 * performs a special pattern match for events
 */
int
match_event (char *mask, char *ev, int cs, int words, int inv, int rest)
{
  char *mask_word = NULL;
  char *mask_list = NULL;
  char *ev_word = NULL;
  int r = 1;

  /* case-sensitive matching */
  for (; r && cs > 0; cs--)
    {
      r = (parse_word (&mask, &mask_word) && parse_word (&ev, &ev_word));
      if (r)
	{
	  char *m = mask_word;
	  r = 0;
	  while (!r && parse_list (&m, &mask_list))
	    r = match (mask_list, ev_word);
	}
    }

  /* word-by-word matching */
  for (; r && words > 0; words--)
    {
      r = (parse_word (&mask, &mask_word) && parse_word (&ev, &ev_word));
      if (r)
	{
	  char *m = mask_word;
	  r = 0;
	  while (!r && parse_list (&m, &mask_list))
	    r = match_irc (mask_list, ev_word);
	}
    }

  /* inverse matching */
  for (; r && inv > 0; inv--)
    {
      r = (parse_word (&mask, &mask_word) && parse_word (&ev, &ev_word));
      if (r)
	{
	  char *m = mask_word;
	  r = 0;
	  while (!r && parse_list (&m, &mask_list))
	    r = match_irc (ev_word, mask_list);
	}
    }

  /* text matching */
  if (r)
    r = (rest > 0 && match_irc (mask, ev)) || (*mask == 0 && *ev == 0);

  free_save (mask_word);
  free_save (mask_list);
  free_save (ev_word);
  return r;
}


/*
 * sets the parameter variables
 */
void
set_params (char *str)
{
  char *word = NULL;
  char index[128];
  struct sl *var;
  int n;

  /* $0- */
  var = var_get ("0-");
  sl_clear (var);
  sl_add (var, str);

  for (n=0; parse_word (&str, &word); n++)
    {
      /* $n */
      sprintf (index, "%i", n);
      var = var_get (index);
      sl_clear (var);
      sl_add (var, word);

      /* $n- */
      sprintf (index, "%i-", n+1);
      var = var_get (index);
      sl_clear (var);
      sl_add (var, str);
    }

  free_save (word);
}


/*
 * reacts on an event
 *
 * format of "ev": <words><cs><inv><rest><event>
 *
 * <cs>    is a single character '0'-'9'
 *         (= number of words to match word by word, but case sensitive)
 *
 * <words> is a single character '0'-'9'
 *         (= number of words to match word by word)
 *
 * <inv>   is a single character '0'-'9'
 *         (= number of words to match word by word, but inverse)
 *
 * <rest>  is a single character '0'-'1'
 *         (= whether to match the rest as text or not to match any rest)
 */
void
event (char *ev)
{
  static int depth = 0;
  static int params_depth = 0;
  int own_params_depth = 0;
  int cs, words, inv, rest;
  struct script_it *it = NULL;
  char *instr = NULL;

  /* set depths */
  depth++;
  own_params_depth = ++params_depth;

  /* validate event parameters */
  if (   ev[0] < '0' || ev[0] > '9'
      || ev[1] < '0' || ev[1] > '9'
      || ev[2] < '0' || ev[2] > '9'
      || ev[3] < '0' || ev[3] > '1')
    {
      print2ilog (L_DETAILED, "-(?,?,?,?) %s", ev);
      print2ilog (L_ERROR, "## invalid event parameters");
      return;
    }

  /* extract event parameters */
  cs	= (*ev++) - '0';
  words	= (*ev++) - '0';
  inv	= (*ev++) - '0';
  rest	= (*ev++) - '0';

  print2ilog (L_DETAILED, "-(%i,%i,%i,%i) %s", cs, words, inv, rest, ev);

  /* set parameter variables */
  set_params (ev);

  /* step through all scripts */
  while (script_iterate (&it))
    {
      int cont;
      char *script = it->script;

      /* look for the first matching event */
      for (cont = 1; cont;)
	switch (parse_type (&script, 1))
	  {
	  case P_END:
	    cont = 0;
	    break;

	  case P_EVENT:
	    parse_next (&script, &instr);
	    if (match_event (instr, ev, cs, words, inv, rest))
	      {
		print2ilog (L_DETAILED, "        -> %s", instr);
		cont = 0;
	      }
	    break;

	  default:
	    parse_skip (&script);
	  }

      /* execute all direct following commands */
      for (cont = 1; cont;)
	switch (parse_type (&script, 1))
	  {
	  case P_CONTENT:
	    parse_next (&script, &instr);
	    command (instr);
	    if (params_depth > own_params_depth)
	      {
		params_depth = own_params_depth;
		set_params (ev);
	      }
	    break;

	  case P_DOLLAR:
	    parse_next (&script, &instr);
	    assign (instr);
	    break;

	  default:
	    cont = 0;
	  }
    }

  /* reset depth */
  depth--;

  /* reload if necessary */
  if (need_reload && depth == 0)
    {
      need_reload = 0;
      reload ();
    }

  free_save (instr);
}


/*
 * reacts on a new IRC server connection
 */
void
event_start (const char *nick, const char *server)
{
  /* create event string */
  char *ev = NULL;
  ev = str_cat (ev, "1200start ");
  ev = str_cat (ev, nick);
  ev = str_cat (ev, " ");
  ev = str_cat (ev, server);

  /* react */
  event (ev);

  free_save (ev);
}


/*
 * reacts on a closed IRC server connection
 */
void
event_quit ()
{
  /* react */
  event ("1000quit");
}


/*
 * takes the CTCP parts out of a message and reacts on them
 */
void
event_ctcp (char **msg, const char *ctcp_start, char *text)
{
  char *ctcp = str_cpy (NULL, ctcp_start);
  int in = 0;
  int msg_len = str_len (*msg);

  for (; *text != 000; text++)
    {
      if (*text == 001)
	{
	  in = !in;
	  if (!in)
	    {
	      event (ctcp);
	      ctcp = str_cpy (ctcp, ctcp_start);
	    }
	}
      else
	{
	  if (in)
	    ctcp = str_catchar (ctcp, *text);
	  else
	    *msg = str_catchar (*msg, *text);
	}
    }

  if (in)
    {
      *msg = str_catchar (*msg, 001);
      *msg = str_cat (*msg, ctcp + str_len (ctcp_start));
    }

  if (str_len (*msg) > msg_len)
    event (*msg);

  free_save (ctcp);
}


/*
 * reacts on a new line from the IRC server
 */
void
event_irc (const char *line)
{
  char *ev = NULL;
  char *c = NULL;
  char *line_copy = str_cpy (NULL, line);

  /* quit <nick> <addr> <text> */
  if (match_irc (":%!% QUIT*", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *tmp2 = split (tmp1, ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1201quit ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, text);
      event (ev);
    }

  /* nick <nick> <addr> <to nick> */
  else if (match_irc (":%!% NICK :%", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *to   = split (tmp1, ' ')+1;

      ev = str_cat (ev, "1300nick ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, to);
      event (ev);
    }

  /* join <nick> <addr> <channel> */
  else if (match_irc (":%!% JOIN :%", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *chan = split (tmp1, ' ')+1;

      ev = str_cat (ev, "1300join ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      event (ev);
    }

  /* part <nick> <addr> <channel> <text> */
  else if (match_irc (":%!% PART *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *chan = split (tmp1, ' ');
      char *tmp2 = split (chan, ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1301part ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, text);
      event (ev);
    }

  /* kick <nick> <addr> <channel> <to nick> <text> */
  else if (match_irc (":%!% KICK % *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *chan = split (tmp1, ' ');
      char *to   = split (chan, ' ');
      char *tmp2 = split (to,   ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1401kick ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, to);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, text);
      event (ev);
    }

  /* msg  <nick> <addr> <channel/nick> <text> */
  /* cmsg <nick> <addr> <channel/nick> <text> */
  else if (match_irc (":%!% PRIVMSG *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *to   = split (tmp1, ' ');
      char *tmp2 = split (to,   ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1301msg ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, to);
      ev = str_cat (ev, " ");

      c = str_cat (c, "1301cmsg ");
      c = str_cat (c, nick);
      c = str_cat (c, " ");
      c = str_cat (c, addr);
      c = str_cat (c, " ");
      c = str_cat (c, to);
      c = str_cat (c, " ");

      event_ctcp (&ev, c, text);
    }

  /* notice  <nick> <addr> <channel/nick> <text> */
  /* cnotice <nick> <addr> <channel/nick> <text> */
  else if (match_irc (":%!% NOTICE *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *to   = split (tmp1, ' ');
      char *tmp2 = split (to,   ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1301notice ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, to);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, text);

      c = str_cat (c, "1301cnotice ");
      c = str_cat (c, nick);
      c = str_cat (c, " ");
      c = str_cat (c, addr);
      c = str_cat (c, " ");
      c = str_cat (c, to);
      c = str_cat (c, " ");

      event_ctcp (&ev, c, text);
    }

  /* topic <nick> <addr> <channel> <text> */
  else if (match_irc (":%!% TOPIC *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *chan = split (tmp1, ' ');
      char *tmp2 = split (chan, ' ');
      char *text = split (tmp2, ':');

      ev = str_cat (ev, "1301topic ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, text);
      event (ev);
    }

  /* invite <nick> <addr> <channel> */
  else if (match_irc (":%!% INVITE % :%", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *tmp2 = split (tmp1, ' ');
      char *chan = split (tmp2, ' ')+1;

      ev = str_cat (ev, "1300invite ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, addr);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      event (ev);
    }

  /* who <nick> <addr> <channel> <server> <state> <description> */
  else if (match_irc ("% 352 % % % % % % % :% *", line))
    {
      char *tmp1 = line_copy;
      char *tmp2 = split (tmp1, ' ');
      char *tmp3 = split (tmp2, ' ');
      char *chan = split (tmp3, ' ');
      char *user = split (chan, ' ');
      char *host = split (user, ' ');
      char *serv = split (host, ' ');
      char *nick = split (serv, ' ');
      char *stat = split (nick, ' ');
      char *tmp4 = split (stat, ' ');
      char *desc = split (tmp4, ' ');

      ev = str_cat (ev, "1501who ");
      ev = str_cat (ev, nick);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, user);
      ev = str_cat (ev, "@");
      ev = str_cat (ev, host);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, chan);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, serv);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, stat);
      ev = str_cat (ev, " ");
      ev = str_cat (ev, desc);
      event (ev);
    }

  /* who_end <channel/mask> */
  else if (match_irc ("% 315 % *", line))
    {
      char *tmp1 = line_copy;
      char *tmp2 = split (tmp1, ' ');
      char *tmp3 = split (tmp2, ' ');
      char *mask = split (tmp3, ' ');
                   split (mask, ' ');

      ev = str_cat (ev, "1300who_end ");
      ev = str_cat (ev, mask);
      event (ev);
    }

  /* channel modes */
  else if (match_irc (":%!% MODE % *", line))
    {
      char *nick = line_copy+1;
      char *addr = split (nick, '!');
      char *tmp1 = split (addr, ' ');
      char *chan = split (tmp1, ' ');
      char *mode = split (chan, ' ');
      char *rest = split (mode, ' ');
      char state = 000;
      enum {T_NOTHING, T_PARAM_0, T_PARAM_1, T_PARAM_I} type;

      for (; *mode != 000; mode++)
	{
	  type = T_NOTHING;
	  switch (*mode)
	    {
	    case '+':
	    case '-':
	      state = *mode;
	      break;

	    case 'i':
	    case 'm':
	    case 'n':
	    case 'p':
	    case 's':
	    case 't':
	      type = T_PARAM_0;
	      break;

	    case 'k':
	    case 'o':
	    case 'v':
	      type = T_PARAM_1;
	      break;

	    case 'b':
	    case 'e':
	    case 'I':
	      type = T_PARAM_I;
	      break;

	    case 'l':
	      type = state == '+' ? T_PARAM_1 : T_PARAM_0;
	      break;
	    }

	  if (type != T_NOTHING)
	    {
	      ev = str_cpy (ev, (type == T_PARAM_0 ? "1300" :
				 type == T_PARAM_1 ? "1400" :
				 type == T_PARAM_I ? "1410" :
				 ""));
	      ev = str_catchar (ev, state);
	      ev = str_catchar (ev, *mode);
	      ev = str_cat (ev, " ");
	      ev = str_cat (ev, nick);
	      ev = str_cat (ev, " ");
	      ev = str_cat (ev, addr);
	      ev = str_cat (ev, " ");
	      ev = str_cat (ev, chan);
	      if (type == T_PARAM_1)
		{
		  char *param = rest;
		  rest = split (param, ' ');

		  ev = str_cat (ev, " ");
		  ev = str_cat (ev, param);
		}
	      else if (type == T_PARAM_I)
		{
		  char *nick = rest;
		  char *addr = split (nick, '!');
		  rest = split (addr, ' ');

		  ev = str_cat (ev, " ");
		  ev = str_cat (ev, nick);
		  ev = str_cat (ev, " ");
		  ev = str_cat (ev, addr);
		}
	      event (ev);
	    }
	}
    }

  /* irc <text> */
  else
    {
      ev = str_cat (ev, "1001irc ");
      ev = str_cat (ev, line);
      event (ev);
    }

  free_save (ev);
  free_save (c);
  free_save (line_copy);
}


/*
 * reacts on a changed system time
 */
void
event_timer (time_t time)
{
  static char ev[128];

  /* create event string */
  strftime (ev, 127, "1200timer %Y-%m-%d %H:%M:%S", localtime (&time));
  ev[127] = 000;

  /* react */
  event (ev);
}


/*
 * reacts on interrupt
 */
void
event_int ()
{
  /* react */
  event ("1000kill");
}
