#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "tools.h"
#include "data.h"
#include "event.h"
#include "io.h"


/*
 * size for internal static string buffers
 */
#define BUFFER_SIZE 1024


/*
 * maximum sleep time for the timer in microseconds
 */
#define SLEEP_US 100000


/*
 * sleep time between the reconnects in seconds
 */
#define SLEEP_CONNECT 30


/*
 * current internal log file
 */
char *log_file = NULL;


/*
 * current mode of the log file
 */
enum log_mode_t log_mode = -1;


/*
 * current socket
 */
static int sock = -1;
static FILE *fsock = NULL;


/*
 * appends a line to a file
 */
void
print2file (const char *filename,
	    const char *prefix, const char *format, va_list ap)
{
  /* open file for appending */
  FILE *f = fopen (filename, "a");

  /* append the line + EOL to the file */
  if (f != NULL)
    {
      fprintf (f, "%s", prefix);
      vfprintf (f, format, ap);
      fprintf (f, "\n");
      fclose (f);
    }
}


/*
 * prints a line into the log
 */
void
print2log (const char *filename, const char *format, ...)
{
  va_list ap;
  va_start (ap, format);

  print2file (filename, "", format, ap);

  va_end (ap);
}


/*
 * prints a line into the internal log
 */
void
print2ilog (enum log_mode_t mode, const char *format, ...)
{
  if (mode <= log_mode)
    {
      time_t t;
      static char c[128];
      va_list ap;
      va_start (ap, format);

      /* get and format time */
      time (&t);
      strftime (c, 127, "[%Y-%m-%d %H:%M:%S] ", localtime (&t));
      c[127] = 000;

      /* print */
      print2file (log_file, c, format, ap);

      va_end (ap);
    }
}


/*
 * sends a line to the socket
 */
void
print2sock (const char *format, ...)
{
  if (fsock != NULL)
    {
      va_list ap;
      va_start (ap, format);

      vfprintf (fsock, format, ap);
      fprintf (fsock, "\015\012");
      fflush (fsock);

      va_end (ap);
    }
}


/*
 * writes a file
 */
void
writefile (const char *filename, struct sl *lines)
{
  struct sl_it *it = NULL;

  /* open file for overwriting */
  FILE *f = fopen (filename, "w");

  if (f != NULL)
    {
      /* write all lines */
      while (sl_iterate (&it, lines))
	fprintf (f, "%s\n", it->str);
      fclose (f);
    }
}


/*
 * reads a file
 */
char *
readfile (const char *filename)
{
  size_t i;
  FILE *f;
  char *c = NULL;
  static char b[BUFFER_SIZE];

  /* open */
  f = fopen (filename, "r");
  if (f == NULL)
    return c;

  /* read */
  while ( (i = fread (b, 1, BUFFER_SIZE, f)) > 0 )
    c = str_catbuf (c, b, i);

  /* close */
  fclose (f);

  return c;
}


/*
 * closes connections to server
 */
void
conn_close ()
{
  /* close sock */
  close (sock);
  sock = -1;

  /* close fsock */
  if (fsock != NULL)
    fclose (fsock);
  fsock = NULL;
}


/*
 * connects to server
 */
void
conn (const char *server, int port)
{
  struct hostent *he;
  struct sockaddr_in addr;
  int s;

  /* close old socket */
  conn_close ();

  /* create a local socket */
  s = socket (AF_INET, SOCK_STREAM, 0);
  if (s < 0) return;

  /* try to resolve server IP */
  he = gethostbyname (server);
  if (he == NULL) return;

  /* connect to server */
  memcpy (&addr.sin_addr, he->h_addr_list[0], he->h_length);
  addr.sin_family = AF_INET;
  addr.sin_port = htons (port);
  if (connect (s, (struct sockaddr*) &addr, sizeof(addr)) < 0) return;

  /* set new socket */
  sock = s;
  fsock = fdopen (s, "a+");
}


/*
 * timer
 */
void
timer ()
{
  static time_t t, old=0;

  /* if time changed ... */
  if (time (&t) != old)
    {
      /* save currect time */
      old = t;

      /* react */
      event_timer (t);
    }
}


/*
 * sleeps sec seconds and calls the timer
 */
void
sleep_timer (unsigned int sec)
{
  struct timeval tv;
  unsigned int cnt = sec*1000000/SLEEP_US;

  while (cnt--)
    {
      timer();

      tv.tv_sec = 0;
      tv.tv_usec = SLEEP_US;
      select (0, NULL, NULL, NULL, &tv);
    }
}


/*
 * waits for and reads a line from the socket
 * calling timer() during the wait time
 */
int
readline (char **result)
{
  static char s[BUFFER_SIZE];
  static int pos=0;
  int l;
  fd_set fs;
  struct timeval tv;

  /* result */
  char *r = str_cpy (*result, "");

  /* invalid socket */
  if (sock < 0)
    {
      /* clear buffer */
      pos = 0;

      /* return */
      *result = r;
      return 0;
    }

  for (;;)
    {
      int i;

      /* look for the first end of line (012) */
      for (i=0; (i < pos) && (s[i] != 012); i++);

      /* if found ... */
      if (i < pos)
	{
	  /* cut it and remove 015, if present */
	  l = (i>0 && s[i-1]==015) ? i-1 : i;

	  /* append to r */
	  r = str_catbuf (r, s, l);

	  /* remove the line from s */
	  i++;
	  pos -= i;
	  memmove (s, s+i, pos);

	  /* return */
	  *result = r;
	  return 1;
	}

      /* if end of line not found and buffer is full ... */
      else if (pos >= BUFFER_SIZE)
	{
	  /* append to r */
	  r = str_catbuf (r, s, BUFFER_SIZE);

	  /* clear buffer */
	  pos = 0;
	}

      /* call timer() until there is new data at the socket */
      do
	{
	  /* timer call */
	  timer ();

	  /* (re)initialize the timeout and the fd to pay attention to */
	  tv.tv_sec = 0;
	  tv.tv_usec = SLEEP_US;
	  FD_ZERO (&fs);
	  FD_SET (sock, &fs);

	  /* reset errno */
	  errno = 0;
	}
      while (select (sock+1, &fs, NULL, NULL, &tv) == 0 || errno == EINTR);

      /* read from socket */
      i = read (sock, s+pos, BUFFER_SIZE-pos);

      /* if read error ... */
      if (i <= 0)
	{
	  /* clear buffer */
	  pos = 0;

	  /* return */
	  *result = r;
	  return 0;
	}

      /* set new position */
      pos += i;
    }
}


/*
 * creates a random, valid nickname
 */
char *
create_nick ()
{
  static char r[10];
  int i;

  r[0] = 'a'+RAND(26);
  for (i=1; i<9; i++)
    {
      int j = RAND(36);
      r[i] = j<10 ? j+'0' : j+'a'-10;
    }
  r[9] = 000;

  return r;
}


/*
 * signal handler for interrupts
 */
void
handler_int (int sig)
{
  event_int ();
}


/*
 * main loop
 */
void
run ()
{
  char *line = NULL;

  struct sl *servers = NULL;
  struct sl_it *servers_it = NULL;
  char *server = NULL;

  struct sl *nicks = NULL;
  struct sl_it *nicks_it = NULL;
  char *nick = NULL;

  char *serv;
  char *port;
  char *pass;

  char *user;
  char *desc;

  /* mode */
  enum {M_INIT, M_CONNECT, M_RUN} mode = M_INIT;

  /* nick-mode */
  enum {N_LIST, N_RANDOM} nick_mode = N_LIST;

  /* initialization */
  srand ((unsigned int) time (NULL));
  signal (SIGPIPE, SIG_IGN);
  signal (SIGINT, handler_int);
  signal (SIGQUIT, handler_int);
  signal (SIGTERM, handler_int);

  /* loop */
  for (;;)
    {
      /* read error */
      if (!readline (&line))
	{
	  /* make sure the connection is closed */
	  conn_close ();

	  /* if there was a connection before ... */
	  if (mode == M_RUN)
	    {
	      /* quit */
	      event_quit ();
	    }

	  /* if we try to connect the first server ... */
	  if (mode != M_CONNECT)
	    {
	      /* reset server list */
	      servers = var_get ("servers");
	      sl_it_stop (&servers_it);
	    }

	  /* return if there are no servers */
	  if (servers->count == 0)
	    {
	      print2ilog (L_NORMAL, "empty server list");
	      sl_it_stop (&servers_it);
	      sl_it_stop (&nicks_it);
	      free_save (line);
	      return;
	    }

	  /* reset nick mode */
	  nick_mode = N_LIST;

	  /* reset nick list */
	  nicks = var_get ("nicks");
	  sl_it_stop (&nicks_it);

	  /* set mode */
	  mode = M_CONNECT;

	  /* if there isn't a next server ... */
	  if (!sl_iterate (&servers_it, servers))
	    {
	      print2ilog (L_NORMAL, "waiting");

	      /* wait */
	      sleep_timer (SLEEP_CONNECT);

	      /* initialize for next loop */
	      sl_it_stop (&servers_it);
	    }

	  /* if there is a next server ... */
	  else
	    {
	      /* split server and user information */
	      serv = str_cpy (NULL, server = servers_it->str);
	      port = split (serv, ':');
	      pass = split (port, ':');
	      user = sl_cat (NULL, var_get ("user"), ',');
	      desc = split (user, ':');

	      /* connect to server */
	      print2ilog (L_NORMAL, "connecting %s", server);
	      conn (serv, *port == 000 ? 6667 : atoi (port));

	      /* set password */
	      if (*pass != 000)
		print2sock ("PASS :%s", pass);

	      /* set user */
	      print2ilog (L_NORMAL, "-] user  %s", user);
	      print2ilog (L_NORMAL, "-] desc  %s", desc);
	      print2sock ("USER %s _ _ :%s",
			  *user == 0 ? "x" : user,
			  *desc == 0 ? "x" : desc);

	      /* set next nick */ 
	      if (nick_mode == N_LIST && !sl_iterate (&nicks_it, nicks))
		nick_mode = N_RANDOM;
	      nick = nick_mode == N_LIST ? nicks_it->str : create_nick ();
	      print2ilog (L_NORMAL, "-] nick  %s", nick);
	      print2sock ("NICK %s", nick);

	      /* destroy split information */
	      free_save (serv);
	      free_save (user);
	    }
	}

      /* login failure */
      else if (match_irc ("ERROR *", line)
	       || match_irc ("% 463 *", line)
	       || match_irc ("% 464 *", line)
	       || match_irc ("% 465 *", line))
	{
	  /* don't waste time */
	  print2ilog (L_NORMAL, "-] error %s", line);
	  conn_close ();
	}

      /* problems with the nick during connecting */
      else if (mode == M_CONNECT
	       && (match_irc ("% 43? *", line) || match_irc ("% 461 *", line)))
	{
	  /* set next nick */ 
	  if (nick_mode == N_LIST && !sl_iterate (&nicks_it, nicks))
	    nick_mode = N_RANDOM;
	  nick = nick_mode == N_LIST ? nicks_it->str : create_nick ();
	  print2ilog (L_NORMAL, "-] nick  %s", nick, line);
	  print2sock ("NICK %s", nick);
	}

      /* ping */
      else if (match_irc ("PING *", line))
	{
	  /* pong */
	  line[1] = 'O';
	  print2sock ("%s", line);
	}

      /* anything else */
      else
	{
	  /* if it was the first time after connecting ... */
	  if (mode == M_CONNECT)
	    {
	      /* start */
	      event_start (nick, server);
	    }

	  /* set mode */
	  mode = M_RUN;

	  /* react */
	  event_irc (line);
	}
    }
}
