#include <config.h>
#include <stdlib.h>
#include <unistd.h>
#include "tools.h"
#include "io.h"
#include "data.h"
#include "parse.h"
#include "loadsave.h"


/*
 * current main file
 */
char *main_file = NULL;


/*
 * current default ini file
 */
char *ini_file = NULL;


/*
 * free internally used memory
 */
void
load_done ()
{
  free_save (main_file);
}


/*
 * set the main file
 */
void
set_main_file (char *filename)
{
  /* split */
  char *dir = NULL;
  char *name = NULL;
  char *suffix = NULL;
  split_path (filename, &dir, &name, &suffix);

  /* change directory */
  chdir (dir);

  /* main file */
  main_file = str_cpy (main_file, name);
  main_file = str_cat (main_file, suffix);

  /* ini file */
  ini_file = str_cpy (ini_file, name);
  ini_file = str_cat (ini_file, ".ini");

  free_save (dir);
}


/*
 * reloads everything
 */
void
reload ()
{
  struct sl *var;

  /* destroy all data */
  data_done();
  data_init();

  /* set default ini */
  var = var_get ("ini");
  sl_clear (var);
  sl_add (var, ini_file);

  /* load main and ini script */
  load (main_file);
  load (sl_get (var_get ("ini"), 0));
}


/*
 * loads a script
 */
void
load (const char *filename)
{
  char *c = NULL;
  char *r = NULL;
  char *t = NULL;
  char *s;
  char *u;
  struct sl *var = NULL;
  int cont;

  /* read script from file */
  c = readfile (filename);

  /* if the file couldn't be read ... */
  if (c == NULL)
    {
      if (filename == NULL || *filename == 000)
	print2ilog (L_ERROR, "## could not load <???>  (invalid file name)");
      else
	print2ilog (L_ERROR, "## could not load <%s>", filename);
      return;
    }

  print2ilog (L_NORMAL, "loading <%s>", filename);

  /* mark as loaded */
  script_put (filename, "");

  /* load variables */
  s = c;
  for (cont = 1; cont;)
    {
      switch (parse_type (&s, 0))
	{
	case P_VAR:
	  /* set current variable */
	  parse_next (&s, &r);
	  u = r;
	  print2ilog (L_DETAILED, "-> %s", u);
	  var = parse_word (&u, &t) ? var_get (t) : NULL;
	  if (var == NULL)
	    print2ilog (L_ERROR, "## invalid variable");

	  /* add to current variable */
	  sl_clear (var);
	  while (parse_list (&u, &t))
	    sl_add_u (var, t);
	  break;

	case P_CONTENT:
	  /* add to current variable */
	  if (var != NULL)
	    {
	      parse_next (&s, &r);
	      u = r;
	      print2ilog (L_DETAILED, "   -> %s", u);
	      while (parse_list (&u, &t))
		sl_add_u (var, t);
	    }
	  else
	    {
	      parse_skip (&s);
	    }
	  break;

	case P_INCLUDE:
	  /* clear current variable */
	  var = NULL;

	  /* retrieve complete file name */
	  parse_next (&s, &r);
	  u = r;
	  print2ilog (L_DETAILED, "-] %s", u);
	  parse_word (&u, &t);

	  /* load if necessary */
	  if (script_get (t) == NULL) load (t);
	  break;

	case P_END:
	  cont = 0;
	  break;

	default:
	  /* clear current variable */
	  var = NULL;

	  /* skip */
	  parse_skip (&s);
	}
    }

  /* insert */
  script_put (filename, c);

  free_save (c);
  free_save (r);
  free_save (t);
}


/*
 * saves all variables
 */
void
save (const char *filename)
{
  char *s = NULL;
  struct sl *l;
  struct sl_it *it = NULL;
  struct var_it *vars = NULL;

  /* check filename */
  if (filename == NULL)
    {
      print2ilog (L_ERROR, "## could not save to <???>  (invalid file name)",
		  filename);
      return;
    }

  print2ilog (L_NORMAL, "saving to <%s>", filename);

  l = sl_new ();

  /* create start */
  sl_add (l, "##");
  sl_add (l, "##  This file was created automatically with ");
  sl_add (l, "##");
  sl_add (l, "##       << Davin - The DAVIS Interpreter - " VERSION " >>");
  sl_add (l, "##");
  s = str_cpy (s, "##  DAVIS ini file \"");
  s = str_cat (s, filename);
  s = str_cat (s, "\"");
  sl_add (l, s);
  sl_add (l, "##");

  /* create content */
  while (var_iterate (&vars))
    {
      if ((vars->name[0] < '0' || vars->name[0] > '9')
	  && (vars->var->count > 0))
	{
	  sl_add (l, "");

	  s = str_cpy (s, "$");
	  quote (vars->name, &s);
	  sl_add (l, s);

	  while (sl_iterate (&it, vars->var))
	    {
	      s = str_cpy (s, "\t[");
	      quote (it->str, &s);
	      s = str_catchar (s, ']');
	      sl_add (l, s);
	    }
	}
    }

  /* create end */
  sl_add (l, "");
  s = str_cpy (s, "## end of ");
  s = str_cat (s, filename);
  sl_add (l, s);

  /* write to ini file */
  writefile (sl_get (var_get ("ini"), 0), l);

  sl_clear (l);
  free_save (l);
  free_save (s);
}
