#include <config.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include "tools.h"
#include "data.h"
#include "io.h"
#include "parse.h"


/*
 * examines whether a character may appear in a name
 */
int
is_name (char ch)
{
  return
       (ch >= 'A' && ch <= 'Z')
    || (ch >= 'a' && ch <= 'z')
    || (ch >= '0' && ch <= '9')
    || (ch == '_')
    || (ch == '-');
}


/*
 * examines whether the end of the instruction is reached
 * if found, let (*script) point to the next instruction
 */
int
parse_reached_end (char **script)
{
  /* script */
  char *s = *script;

  /* skip leading spaces */
  while (is_space (*s)) s++;

  /* skip comments */
  if (*s == '#' && s[1] == '#')
    {
      while (!is_newline (*s) && *s != 000) s++;
    }

  /* reached end? */
  if (is_newline (*s))
    {
      *script = s+1;
      return 1;
    }
  if (*s == 000)
    {
      *script = s;
      return 1;
    }
  return 0;
}


/*
 * parses a string in math mode
 */
pm_number
parse_math (char **script, char ch_close, int fast)
{
  /* script */
  char *s = *script;

  /* result */
  pm_number r = 0;

  /* temporary strings */
  char *t = NULL;

  /* parse loop */
  int end = 0;
  while (!end && !parse_reached_end (&s))
    {
      switch (*s)
	{
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	  /* digit */
	  r = (10*r) + (*s-'0');
	  s++;
	  break;

	case '+':
	  /* add */
	  s++;
	  r += parse_math (&s, ch_close, fast);
	  end = 1;
	  break;

	case '-':
	  /* subtract */
	  s++;
	  r -= parse_math (&s, ch_close, fast);
	  end = 1;
	  break;

	case '*':
	  /* multiply */
	  s++;
	  r *= parse_math (&s, ch_close, fast);
	  end = 1;
	  break;

	case '/':
	  /* divide */
	  s++;
	  r = (pm_number) ((double)r / parse_math (&s, ch_close, fast) + 0.5);
	  end = 1;
	  break;

	case '^':
	  /* power */
	  s++;
	  r = (pm_number) (pow (r, parse_math (&s, ch_close, fast)) + 0.5);
	  end = 1;
	  break;

	case '(':
	  /* sub statement () */
	  s++;
	  r = parse_math (&s, ')', fast);
	  break;

	case '[':
	  /* sub statement [] */
	  s++;
	  r = parse_math (&s, ']', fast);
	  break;

	case '{':
	  /* sub statement {} */
	  s++;
	  r = parse_math (&s, '}', fast);
	  break;

	case '$':
	  /* variable */
	  {
	    struct sl *var;
	    struct sl_it *i = NULL;
	    enum {O_NORMAL, O_COUNT, O_ADD, O_MUL} operation;
	    s++;

	    /* prefix */
	    switch (*s)
	      {
	      case '#':
		s++;
		operation = O_COUNT;
		break;

	      case '+':
		s++;
		operation = O_ADD;
		break;

	      case '*':
		s++;
		operation = O_MUL;
		break;

	      default:
		operation = O_NORMAL;
	      }

	    /* name */
	    t = str_cpy (t, "");
	    if (*s == '[')
	      {
		s++;
		parse_string (&s, &t, 1, fast);
	      }
	    else
	      {
		while (is_name (*s))
		  {
		    str_catchar (t, *s);
		    s++;
		  }
	      }

	    if (fast) break;
	    var = var_get (t);

	    /* calculate */
	    switch (operation)
	      {
	      case O_COUNT:
		r = (pm_number) var->count;
		break;

	      case O_ADD:
		r = 0;
		while (sl_iterate (&i, var)) r += (pm_number) atol (i->str);
		break;

	      case O_MUL:
		r = 1;
		while (sl_iterate (&i, var)) r *= (pm_number) atol (i->str);
		break;

	      default:
		r = (pm_number) (var->count == 0
				 ? 0 : atol (sl_get (var, 0)));
	      }
	  }
	  break;

	default:
	  /* end of statement */
	  if (*s == ch_close)
	    {
	      s++;
	      end = 1;
	    }

	  /* unknown character */
	  else
	    {
	      s++;
	    }
	}
    }

  free_save (t);
  *script = s;
  return r;
}


/*
 * parses a file
 * appends the result to *result
 */
void
parse_file (char *file, char **result)
{
  int first = 0;
  int begin = 1;
  for (; *file != 000; file++)
    {
      if (is_newline (*file))
	{
	  first = 1;
	}
      else
	{
	  if (first)
	    {
	      if (!begin)
		*result = str_catchar (*result, ',');
	      first = 0;
	    }
	  *result = str_catchar (*result, *file);
	  begin = 0;
	}
    }  
}


/*
 * parses a string
 * appends the result to *result
 */
void
parse_string (char **script, char **result, int sub, int fast)
{
  /* script */
  char *s = *script;

  /* result */
  char *r = str_cpy (NULL, "");

  /* temporary strings */
  char *t = NULL;
  char *u = NULL;
  char *i1 = NULL;
  char *i2 = NULL;

  /* parse loop */
  int end = 0;
  while (!end && !parse_reached_end (&s))
    {
      switch (*s)
	{
	case ']':
	  /* end of a sub parse */
	  if (sub)
	    {
	      s++;
	      end = 1;
	    }
	  break;

	case '\t':
	  /* one or more tabs -> one space */
	  while (*s == '\t') s++;
	  r = str_catchar (r, ' ');
	  break;

	case '\\':
	  /* escape sequences */
	  s++;
	  switch (*s)
	    {
	    case '\\':
	    case '$':
	    case '#':
	    case '[':
	    case ']':
	    case '|':
	      /* special character */
	      r = str_catchar (r, *s);
	      s++;
	      break;

	    case 'i':
	      /* ignore */
	      s++;
	      break;

	    case 't':
	      /* tabulator */
	      s++;
	      r = str_catchar (r, 007);
	      break;

	    case 'b':
	      /* bold */
	      s++;
	      r = str_catchar (r, 002);
	      break;

	    case 'c':
	      /* CTCP */
	      s++;
	      r = str_catchar (r, 001);
	      break;

	    default:
	      /* ignore EOL */
	      if (parse_reached_end (&s))
		{
		  while (*s != 000 && parse_reached_end (&s));
		  while (is_space (*s)) s++;
		}
	    }
	  break;

	case '[':
	  /* sub parse */
	  s++;
	  parse_string (&s, &r, 1, fast);
	  break;

	case '|':
	  /* combine */
	  {
	    int first;
	    char *t2, *u2;

	    s++;
	    if (fast) break;

	    t = str_cpy (t, r);
	    u = str_cpy (u, "");
	    parse_string (&s, &u, sub, fast);
	    r = str_cpy (r, "");

	    first = 1;
	    for (t2 = t; parse_list (&t2, &i1);)
	      for (u2 = u; parse_list (&u2, &i2);)
		{
		  if (first)
		    first = 0;
		  else
		    r = str_catchar (r, ',');

		  r = str_cat (r, i1);
		  r = str_cat (r, i2);
		}
	  }
	  end = 1;
	  break;

	case '$':
	  /* variable */
	  {
	    s++;

	    /* name */
	    t = str_cpy (t, "");
	    if (*s == '[')
	      {
		s++;
		parse_string (&s, &t, 1, fast);
	      }
	    else
	      {
		while (is_name (*s))
		  {
		    str_catchar (t, *s);
		    s++;
		  }
	      }

	    /* suffix */
	    switch (*s)
	      {
	      case '%':
		/* random */
		{
		  struct sl *var;
		  struct sl *l;
		  int i = INT_MAX;
		  s++;

		  if (is_digit (*s))
		    {
		      for (i = 0; is_digit (*s); s++)
			i = (10*i) + (*s-'0');
		    }

		  if (fast) break;
		  var = var_get (t);
		  l = sl_new ();

		  while (l->count < var->count && l->count < i)
		    {
		      char *s = sl_rand (var);
		      sl_add_u (l, s);
		    }

		  r = sl_cat (r, l, ',');

		  sl_clear (l);
		  free_save (l);
		  break;
		}

	      default:
		/* normal */
		if (fast) break;
		r = sl_cat (r, var_get (t), ',');
	      }
	  }
	  break;

	case '=':
	  /* math mode */
	  if (s[1] != '[')
	    {
	      r = str_catchar (r, *s);
	      s++;
	      break;
	    }

	  s += 2;
	  str_catnr (r, parse_math (&s, ']', fast));
	  break;

	case '&':
	  /* file access */
	  {
	    char *f;

	    if (s[1] != '[')
	      {
		r = str_catchar (r, *s);
		s++;
		break;
	      }

	    s += 2;
	    t = str_cpy (t, "");
	    parse_string (&s, &t, 1, fast);

	    if (fast) break;
	    f = readfile (t);
	    if (f == NULL)
	      {
		print2ilog (L_ERROR, "## could not read &[%s]", t);
		break;
	      }

	    print2ilog (L_NORMAL, "reading &[%s]", t);
	    parse_file (f, &r);
	    free_save (f);
	  }
	  break;

	case '@':
	  /* web access */
	  {
	    char *f;

	    if (s[1] != '[')
	      {
		r = str_catchar (r, *s);
		s++;
		break;
	      }

	    s += 2;
	    t = str_cpy (t, "");
	    parse_string (&s, &t, 1, fast);

	    if (fast) break;
	    f = readurl (t);
	    if (f == NULL)
	      {
		print2ilog (L_ERROR, "## could not read @[%s]", t);
		break;
	      }

	    print2ilog (L_NORMAL, "reading @[%s]", t);
	    parse_file (f, &r);
	    free_save (f);
	  }
	  break;

	default:
	  /* normal character */
	  r = str_catchar (r, *s);
	  s++;
	}
    }

  free_save (t);
  free_save (u);
  free_save (i1);
  free_save (i2);
  *script = s;
  *result = str_cat (*result, r);
  free_save (r);
}


/*
 * replaces the special characters of str with escape sequences
 */
void
quote (const char *str, char **result)
{
  for (; *str != 000; str++)
    {
      switch (*str)
	{
	case 007:
	  *result = str_cat (*result, "\\t");
	  break;

	case 002:
	  *result = str_cat (*result, "\\b");
	  break;

	case 001:
	  *result = str_cat (*result, "\\c");
	  break;

	case '\\':
	case '$':
	case '#':
	case '[':
	case ']':
	case '|':
	  *result = str_catchar (*result, '\\');

	default:
	  *result = str_catchar (*result, *str);
	}
    }
}


/*
 * goes to the next instruction ans returns the type of it
 */
char
parse_type (char **script, int parse_dollar)
{
  /* skip empty instructions */
  while (**script != 000 && parse_reached_end (script));

  /* get type */
  switch (**script)
    {
    case 000:
      return P_END;

    case '?':
      (*script)++;
      return P_INCLUDE;

    case '$':
      (*script)++;
      return P_VAR;

    default:
      if (is_space (**script))
	{
	  while (is_space (**script)) (*script)++;

	  if (parse_dollar && **script == '$')
	    {
	      (*script)++;
	      return P_DOLLAR;
	    }
	  return P_CONTENT;
	}
      return P_EVENT;
    }
}


/*
 * gets the next instruction
 */
void
parse_next (char **script, char **result)
{
  /* initialize result */
  *result = str_cpy (*result, "");

  /* parse */
  parse_string (script, result, 0, 0);
}


/*
 * skips the next instruction
 */
void
parse_skip (char **script)
{
  /* initialize result */
  char *result = NULL;

  /* parse */
  parse_string (script, &result, 0, 1);

  /* forget result */
  free_save (result);
}


/*
 * gets the next word
 */
int
parse_word (char **line, char **result)
{
  /* result */
  char *r = str_cpy (*result, "");

  /* line */
  char *s = *line;

  /* skip spaces */
  while (is_space (*s)) s++;

  /* get the word */
  while (*s != 000 && !is_space (*s))
    {
      r = str_catchar (r, *s);
      s++;
    }

  /* skip spaces */
  while (is_space (*s)) s++;

  /* return */
  *line = s;
  *result = r;
  return *r != 000;
}


/*
 * gets the next element of a comma separated list
 */
int
parse_list (char **list, char **result)
{
  char *r;

  /* list */
  char *s = *list;

  /* no more elements ... */
  if (*s == 000)
    return 0;

  /* result */
  r = str_cpy (*result, "");

  /* get the element */
  while (*s != 000 && *s != ',')
    {
      r = str_catchar (r, *s);
      s++;
    }

  /* skip comma */
  if (*s == ',') s++;

  /* return */
  *list = s;
  *result = r;
  return 1;
}
