#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tools.h"


/*
 * examines whether a character is a "space"
 */
int
is_space (char ch)
{
  return ch == ' ' || ch == '\t';
}


/*
 * examines whether a character is a line delimiter
 */
int
is_newline (char ch)
{
  return ch == '\n' || ch == '\r';
}


/*
 * examines whether a character is a valid digit
 */
int
is_digit (char ch)
{
  return (ch >= '0' && ch <= '9');
}


/*
 * a "strlen" that is able to deal with NULL strings
 */
size_t
str_len (const char *str)
{
  return (str == NULL) ? 0 : strlen (str);
}


/*
 * a "free" that checks for NULL
 */
void
free_save (void *ptr)
{
  if (ptr != NULL)
    free (ptr);
}


/*
 * help macro for str_realloc
 */
#define SIZE_ROUND(x) ((((x)+SIZE_INCREMENT-1)/SIZE_INCREMENT)*SIZE_INCREMENT)


/*
 * reallocates a string to have at least the given length
 */
char *
str_realloc (char *dest, size_t len)
{
  /* reallocate if necessary */
  if ( (dest==NULL) || (SIZE_ROUND(len+1) != SIZE_ROUND(str_len (dest) + 1)) )

    dest = (char *) realloc (dest, SIZE_ROUND(len+1));

  return dest;
}


/*
 * copies src to dest and reallocates dest if necessary
 */
char *
str_cpy (char *dest, const char *src)
{
  /* reallocate */
  dest = str_realloc (dest, str_len(src));

  /* copy */
  strcpy (dest, src);

  return dest;
}


/*
 * appends src to dest and reallocates dest if necessary
 */
char *
str_cat (char *dest, const char *src)
{
  if (dest == NULL)
    return str_cpy (dest, src);

  /* reallocate */
  dest = str_realloc (dest, str_len (dest) + str_len (src));

  /* append */
  strcat (dest, src);

  return dest;
}


/*
 * appends a char to dest and reallocates dest if necessary
 */
char *
str_catchar (char *dest, char ch)
{
  size_t l;

  /* ch may not be 000 */
  if (ch == 000)
    return dest;

  /* reallocate */
  l = str_len (dest);
  dest = str_realloc (dest, l+1);

  /* append */
  dest[l] = ch;
  dest[l+1] = 000;

  return dest;
}


/*
 * replaces zeros in buf, appends buf to dest
 * and reallocates dest if necessary
 */
char *
str_catbuf (char *dest, const void *buf, size_t size)
{
  int i, l;

  /* get length of dest */
  l = str_len (dest);

  /* reallocate */
  dest = str_realloc(dest, l+size);

  /* append and replace zeros */
  for (i=0; i < size; i++)
    dest[l+i] = ((char *)buf)[i] ? ((char *)buf)[i] : ' ';

  dest[l+size] = 0;

  return dest;
}


/*
 * appends an int to dest and reallocates dest if necessary
 */
char *
str_catnr (char *dest, long int nr)
{
  char s[128];
  sprintf (s, "%li", nr);
  return str_cat (dest, s);
}


/*
 * splits str at delim:
 * The first part remains in str, the second part is returned.
 */
char *
split (char *str, char delim)
{
  /* search the delimiter */
  while (*str != 000 && *str != delim) str++;

  /* insert a 000 if necessary */
  if (*str != 000) *str++ = 000;

  /* return */
  return str;
}


/*
 * splits path into directory, filename and suffix.
 * dir, name, suffix may be NULL if their values aren't needed.
 */
void
split_path (char *path, char **dir, char **name, char **suffix)
{
  char *d, *s;

  /* search the directory delimiter */
  d = strrchr (path, '/');
  if (d == NULL) d = strrchr (path, '\\');
  if (d == NULL) d = path - 1;
  d++;

  /* search the suffix delimiter */
  s = strchr (d, '.');
  if (s == NULL) s = d + str_len (d);

  /* copy parts */
  if (dir != NULL)
    {
      *dir = str_cpy (*dir, "");
      *dir = str_catbuf (*dir, path, d - path);
    }
  if (name != NULL)
    {
      *name = str_cpy (*name, "");
      *name = str_catbuf (*name, d, s - d);
    }
  if (suffix != NULL)
    {
      *suffix = str_cpy (*suffix, s);
    }
}


/*
 * performs a simple pattern match
 */
int
match (const char *mask, const char *str)
{
  switch (*mask)
    {
    case 000:
      return *str == 000;

    case '?':
      return (*str != 000) && match (mask+1, str+1);

    case '*':
      mask++;
      do
	{
	  if (match (mask, str))
	    return 1;
	}
      while (*str++ != 000);
      return 0;

    case '%':
      mask++;
      do
	{
	  if (match (mask, str))
	    return 1;
	}
      while ((*str != 000) && (*str != ' ') && (*str++ != 007));
      return 0;

    default:
      return (*str == *mask) && match (mask+1, str+1);
    }
}


/*
 * converts a character to its lower case equivalent
 * as defined in the IRC Protocol (RFC 1459)
 */
char
tolower_irc (char ch)
{
  if (ch >= 'A' && ch <= 'Z') return ch - 'A' + 'a';

  switch (ch)
    {
    case '[':  return '{';
    case ']':  return '}';
    case '\\': return '|';
    }

  return ch;
}


/*
 * converts a string to its lower case equivalent using tolower_irc()
 */
char *
str_tolower_irc (char *dest, const char *src)
{
  char *t;

  /* reallocate */
  dest = str_realloc (dest, str_len (src));

  /* convert */
  for (t = dest; *src != 000; src++, t++)
    *t = tolower_irc (*src);
  *t = 000;

  return dest;
}


/*
 * performs a simple pattern match for IRC strings
 */
int
match_irc (const char *mask, const char *str)
{
  char *m = str_tolower_irc (NULL, mask);
  char *s = str_tolower_irc (NULL, str);

  int r = match (m, s);

  free_save (m);
  free_save (s);
  return r;
}


/*
 * creates a new string list
 */
struct sl *
sl_new ()
{
  /* allocate memory */
  struct sl *list = (struct sl *) malloc (sizeof(struct sl));

  /* initialize */
  list->count = 0;
  list->last = NULL;

  return list;
}


/*
 * iterates a string list
 */
int
sl_iterate (struct sl_it **it, struct sl *list)
{
  /* empty list ... */
  if (list->last == NULL)
    return 0;

  /* first element ... */
  if (*it == NULL)
    {
      /* allocate memory */
      *it = (struct sl_it *) malloc (sizeof(struct sl_it));

      /* select first element */
      (*it)->current = list->last->next;
    }

  /* last element ... */
  else if ((*it)->current == list->last)
    {
      /* free iterator */
      free_save (*it);
      *it = NULL;

      return 0;
    }

  /* any element between ... */
  else
    {
      /* select next element */
      (*it)->current = (*it)->current->next;
    }

  /* copy data */
  (*it)->str = (*it)->current->str;

  return 1;
}


/*
 * stops an iteration
 */
void
sl_it_stop (struct sl_it **it)
{
  /* free iterator if necessary */
  if (*it != NULL)
    {
      /* free iterator */
      free_save (*it);
      *it = NULL;
    }
}


/*
 * finds the node *before* a special string
 */
struct sl_node *
sl_find (struct sl *list, const char *str)
{
  /* if list is non-empty ... */
  if (list->last != NULL)
    {
      /* search */
      struct sl_node *t = list->last;

      do
	{
	  /* if found, return it */
	  if (strcmp (t->next->str, str) == 0) return t;

	  /* next node */
	  t = t->next;
	}
      while (t != list->last);
    }

  /* if list is empty or string not found ... */
  return NULL;
}


/*
 * finds the node *before* a special string using match()
 */
struct sl_node *
sl_find_m (struct sl *list, const char *mask)
{
  /* if list is non-empty ... */
  if (list->last != NULL)
    {
      /* search */
      struct sl_node *t = list->last;

      do
	{
	  /* if found, return it */
	  if (match (mask, t->next->str)) return t;

	  /* next node */
	  t = t->next;
	}
      while (t != list->last);
    }

  /* if list is empty or string not found ... */
  return NULL;
}


/*
 * returns the index'th string of a list
 */
char *
sl_get (struct sl *list, int index)
{
  struct sl_node *t;

  /* invalid index */
  if ((index < 0) || (index >= list->count))
    return NULL;

  /* step through the nodes */
  t = list->last->next;
  while (index--)
    t = t->next;

  return t->str;
}


/*
 * returns a random string of a list
 */
char *
sl_rand (struct sl *list)
{
  /* return the element at a random index */
  return sl_get (list, RAND(list->count));
}


/*
 * appends a string list as delim-separated list to dest
 */
char *
sl_cat (char *dest, struct sl *list, char delim)
{
  struct sl_it *it = NULL;

  if (sl_iterate (&it, list))
    {
      dest = str_cat (dest, it->str);
      while (sl_iterate (&it, list))
	{
	  dest = str_catchar (dest, delim);
	  dest = str_cat (dest, it->str);
	}
    }
  else
    {
      dest = str_cat (dest, "");
    }

  return dest;
}


/*
 * adds a string to a list
 */
void
sl_add (struct sl *list, const char *str)
{
  /* allocate memory for new node */
  struct sl_node *t = (struct sl_node *) malloc (sizeof(struct sl_node));

  /* initialize string */
  t->str = str_cpy (NULL, str);

  /* relink */
  if (list->last != NULL)
    {
      t->next = list->last->next;
      list->last->next = t;
    }
  else
    t->next = t;

  /* reinitialize last */
  list->last = t;

  /* increase count */
  list->count++;
}


/*
 * adds a string to a list if isn't already contained
 */
void
sl_add_u (struct sl *list, const char *str)
{
  if (sl_find (list, str) == NULL)
    sl_add (list, str);
}


/*
 * removes a single node from a list
 * NOTE: the argument has to refer to the node *before* the node to remove
 */
void
sl_remove_node (struct sl *list, struct sl_node *prev)
{
  /* save the node to remove */
  struct sl_node *t = prev->next;

  /* relink */
  prev->next = t->next;

  /* reinitialize "last" */
  list->last = ((prev==t)       ? NULL :
		(list->last==t) ? prev :
				  list->last);

  /* free string and node */
  free_save (t->str);
  free_save (t);

  /* decrease count */
  list->count--;
}


/*
 * removes a string from a list
 */
void
sl_remove (struct sl *list, const char *str)
{
  /* search it */
  struct sl_node *t = sl_find (list, str);
  if (t == NULL) return;

  /* remove it */
  sl_remove_node (list, t);
}


/*
 * removes all strings matching a certain pattern from a list
 */
void
sl_remove_m (struct sl *list, const char *mask)
{
  for (;;)
    {
      /* search it */
      struct sl_node *t = sl_find_m (list, mask);
      if (t == NULL) return;

      /* remove it */
      sl_remove_node (list, t);
    }
}


/*
 * removes all strings from a list
 */
void
sl_clear (struct sl *list)
{
  /* if the list isn't empty yet */
  if (list->last != NULL)
    {
      /* step through the list */
      struct sl_node *t=list->last;

      do
	{
	  /* save current node */
	  struct sl_node *u = t;

	  /* next node */
	  t = t->next;

	  /* free string and node */
	  free_save (u->str);
	  free_save (u);
	}
      while (t != list->last);

      /* initialize */
      list->count = 0;
      list->last = NULL;
    }
}


/*
 * copies a string list
 */
struct sl *
sl_copy (struct sl *list)
{
  struct sl *r = sl_new ();
  struct sl_it *it = NULL;

  while (sl_iterate (&it, list))
    {
      sl_add_u (r, it->str);
    }

  return r;
}


/*
 * creates a new hash table
 */
struct hash *
hash_new ()
{
  /* allocate memory */
  struct hash *table = (struct hash *) malloc (sizeof(struct hash));

  /* initialize */
  table->first = NULL;

  return table;
}


/*
 * iterates a hash table
 */
int
hash_iterate (struct hash_it **it, struct hash *table)
{
  /* empty hash table ... */
  if (table->first == NULL)
    return 0;

  /* first element ... */
  if (*it == NULL)
    {
      /* allocate memory */
      *it = (struct hash_it *) malloc (sizeof(struct hash_it));

      /* select first element */
      (*it)->current = table->first;
    }

  /* last element ... */
  else if ((*it)->current->next == NULL)
    {
      /* free iterator */
      free_save (*it);
      *it = NULL;

      return 0;
    }

  /* any element between ... */
  else
    {
      /* select next element */
      (*it)->current = (*it)->current->next;
    }

  /* if the value is invalid ... */
  if ((*it)->current->value == NULL)
    return hash_iterate (it, table);

  /* copy data */
  (*it)->key = (*it)->current->key;
  (*it)->value = (*it)->current->value;

  return 1;
}


/*
 * stops an iteration
 */
void
hash_it_stop (struct hash_it **it)
{
  /* free iterator if necessary */
  if (*it != NULL)
    {
      /* free iterator */
      free_save (*it);
      *it = NULL;
    }
}


/*
 * returns the value of a key
 */
void *
hash_get (struct hash *table, const char *key)
{
  /* search */
  struct hash_node *t = table->first;

  while (t)
    {
      /* if found ... */
      if (strcmp (key, t->key) == 0)
	return t->value;

      /* next node */
      t = t->next;
    }

  /* if not found ... */
  return NULL;
}


/*
 * sets the value of a key and returns the old value
 */
void *
hash_put (struct hash *table, const char *key, void *value)
{
  /* search */
  struct hash_node *t = table->first;

  while (t)
    {
      /* if found ... */
      if (strcmp (key, t->key) == 0)
	{
	  /* save old value */
	  void *v = t->value;

	  /* set new value */
	  t->value = value;

	  /* return old value */
	  return v;
	}

      /* next node */
      t = t->next;
    }

  /* if not found ... */

  /* create a new node */
  t = (struct hash_node *) malloc (sizeof(struct hash_node));

  /* initialize */
  t->key = str_cpy (NULL, key);
  t->value = value;

  /* relink */
  t->next = table->first;
  table->first = t;

  return NULL;
}


/*
 * removes all key-value pairs from a hash table
 */
void
hash_clear (struct hash *table)
{
  /* step through the hash table */
  struct hash_node *t=table->first;

  while (t)
    {
      /* save current node */
      struct hash_node *u = t;

      /* next node */
      t = t->next;

      /* free key and node */
      free_save (u->key);
      free_save (u);

    }

  /* initialize */
  table->first = NULL;
}
