#ifndef __TOOLS_HEADER
#define __TOOLS_HEADER


#include <config.h>


/*
 * examines whether a character is a "space"
 */
int
is_space (char ch);


/*
 * examines whether a character is a line delimiter
 */
int
is_newline (char ch);


/*
 * examines whether a character is a valid digit
 */
int
is_digit (char ch);


/*
 * size increment for dynamic string buffers (4k)
 */
#define SIZE_INCREMENT (4*1024)


/*
 * returns a random value in the range of 0 ... (m-1)
 */
#define RAND(m) (int) ( (double)rand() * (m) / RAND_MAX )


/*
 * a "free" that checks for NULL
 */
void
free_save (void *ptr);


/*
 * a "strlen" that is able to deal with NULL strings
 */
size_t
str_len (const char *str);


/*
 * copies src to dest and reallocates dest if necessary
 */
char *
str_cpy (char *dest, const char *src);


/*
 * appends src to dest and reallocates dest if necessary
 */
char *
str_cat (char *dest, const char *src);


/*
 * appends a char to dest and reallocates dest if necessary
 */
char *
str_catchar (char *dest, char ch);


/*
 * replaces zeros in buf, appends buf to dest
 * and reallocates dest if necessary
 */
char *
str_catbuf (char *dest, const void *buf, size_t size);


/*
 * appends an int to dest and reallocates dest if necessary
 */
char *
str_catnr (char *dest, long int nr);


/*
 * splits str at delim:
 * The first part remains in str, the second part is returned.
 */
char *
split (char *str, char delim);


/*
 * splits path into directory, filename and suffix.
 * dir, name, suffix may be NULL if their values aren't needed.
 */
void
split_path (char *path, char **dir, char **name, char **suffix);


/*
 * performs a simple pattern match
 */
int
match (const char *mask, const char *str);


/*
 * converts a character its lower case equivalent
 * as defined in the IRC Protocol (RFC 1459)
 */
char
tolower_irc (char ch);


/*
 * converts a string to its lower case equivalent using tolower_irc()
 */
char *
str_tolower_irc (char *dest, const char *src);


/*
 * performs a simple pattern match for IRC strings
 */
int
match_irc (const char *mask, const char *str);


/*
 * string list
 */
struct sl
{
  int count;
  struct sl_node *last;
};


/*
 * node for string lists
 */
struct sl_node
{
  char *str;
  struct sl_node *next;
};


/*
 * iterator for string lists
 */
struct sl_it
{
  char *str;
  struct sl_node *current;
};


/*
 * creates a new string list
 */
struct sl *
sl_new ();


/*
 * iterates a string list
 */
int
sl_iterate (struct sl_it **it, struct sl *list);


/*
 * stops an iteration
 */
void
sl_it_stop (struct sl_it **it);


/*
 * finds the node *before* a special string
 */
struct sl_node *
sl_find (struct sl *list, const char *str);


/*
 * finds the node *before* a special string using match()
 */
struct sl_node *
sl_find_m (struct sl *list, const char *mask);


/*
 * returns the index'th string of a list
 */
char *
sl_get (struct sl *list, int index);


/*
 * returns a random string of a list
 */
char *
sl_rand (struct sl *list);


/*
 * appends a string list as delim-separated list to dest
 */
char *
sl_cat (char *dest, struct sl *list, char delim);


/*
 * adds a string to a list
 */
void
sl_add (struct sl *list, const char *str);


/*
 * adds a string to a list if isn't already contained
 */
void
sl_add_u (struct sl *list, const char *str);


/*
 * removes a single node from a list
 * NOTE: the argument has to refer to the node *before* the node to remove
 */
void
sl_remove_node (struct sl *list, struct sl_node *prev);


/*
 * removes a string from a list
 */
void
sl_remove (struct sl *list, const char *str);


/*
 * removes all strings matching a certain pattern from a list
 */
void
sl_remove_m (struct sl *list, const char *mask);


/*
 * removes all strings from a list
 */
void
sl_clear (struct sl *list);


/*
 * copies a string list
 */
struct sl *
sl_copy (struct sl *list);


/*
 * hash table
 */
struct hash
{
  struct hash_node *first;
};


/*
 * node for hash tables
 */
struct hash_node
{
  char *key;
  void *value;
  struct hash_node *next;
};


/*
 * iterator for hash tables
 */
struct hash_it
{
  char *key;
  void *value;
  struct hash_node *current;
};


/*
 * creates a new hash table
 */
struct hash *
hash_new ();


/*
 * iterates a hash table
 */
int
hash_iterate (struct hash_it **it, struct hash *table);


/*
 * stops an iteration
 */
void
hash_it_stop (struct hash_it **it);


/*
 * returns the value of a key
 */
void *
hash_get (struct hash *table, const char *key);


/*
 * sets the value of a key and returns the old value
 */
void *
hash_put (struct hash *table, const char *key, void *value);


/*
 * removes all key-value pairs from a hash table
 */
void
hash_clear (struct hash *table);


#endif
